import http from 'http';
import https from 'https';
import { Transform, PassThrough } from 'stream';
import zlib from 'zlib';
import typeis from 'type-is';
import destroy from 'destroy';
import { CONTENT_TYPES, brotliOptions, zlibOptions } from './contant';
import { hasbody } from './utils';

export type ReqInfo = {
  remoteAddress: string;
  url: string;
  time: number;
  method: string;
  headers: Record<string, undefined | number | string | string[]>;
  type?: string;
  rawBody?: Buffer;
  textBody?: string;
};

export type ResInfo = {
  cost: number;
  status: number;
  headers: Record<string, undefined | number | string | string[]>;
  type?: string;
  rawBody?: Buffer;
  textBody?: string;
};

export type HttpEvent = {
  type: string;
  req: ReqInfo;
  res: ResInfo;
};

export type HttpListener = (event: HttpEvent) => void;

let currentListener: HttpListener | null = null;

export function inspect(listener: HttpListener | null): void {
  currentListener = listener;
}

export function emit(event: HttpEvent) {
  if (currentListener) {
    currentListener(event);
  }
}

function inspectedMethod(this: any, func: any, ...rest: any) {
  if (!currentListener) {
    return func.apply(this, rest);
  }
  const req: http.ClientRequest = func.apply(this, rest);
  inspectRequest(req);
  return req;
};

function inspectRequest(request: http.ClientRequest) {
  Promise.all([
    collectRemoteAddress(request),
    collectReq(request),
    collectRes(request),
  ]).then(([ remoteAddress, req, res ]) => {

    if (remoteAddress) {
      req.remoteAddress = remoteAddress;
      res.cost = Date.now() - req.time;
    }
    if (req.type && req.rawBody) {
      req.textBody = req.rawBody.toString('utf8')
    }
    emit({ type: 'http/https', req, res });
  }).catch((error) => {
    console.log(error);
  });
}

function collectRemoteAddress(request: http.ClientRequest): Promise<string | null> {
  return new Promise((resolve) => {
    request.prependOnceListener('socket', (socket) => {
      socket.on('connect', () => resolve(socket.remoteAddress ?? null));
      socket.on('error', () => resolve(null));
      socket.on('close', () => resolve(null));
    });
  });
}

function collectReq(request: http.ClientRequest): Promise<ReqInfo> {
  return new Promise((resolve, reject) => {

    const headers = request.getHeaders();
    const req: ReqInfo = {
      time: Date.now(),
      headers: headers,
      type: hasbody(headers)
        ? typeis.is(String(headers['content-type'] ?? ''), CONTENT_TYPES) || undefined
        : undefined,
      method: request.method,
      remoteAddress: '',
      url: `${request.protocol}//${request.getHeader('host')}${request.path}`,
    };

    if (req.type) {

      const requestBody: Buffer[] = [];

      const reqWrite = request.write;
      request.write = function (...args: any) {
        /**
         * chunk can be either a string or a Buffer.
         */
        const [chunk] = args;

        if (Buffer.isBuffer(chunk)) {
          requestBody.push(chunk);
        } else {
          requestBody.push(Buffer.from(chunk, 'utf8'));
        }

        return reqWrite.apply(this, args);
      };

      const reqEnd = request.end;
      request.end = function (...args: any) {
        /**
         * the first argument might be a callback or a chunk
         */
        const [maybeChunk] = args;

        if (Buffer.isBuffer(maybeChunk)) {
          requestBody.push(maybeChunk);
        } else if (maybeChunk && typeof maybeChunk !== 'function') {
          requestBody.push(Buffer.from(maybeChunk, 'utf8'));
        }

        return reqEnd.apply(this, args);
      };

      request.once('finish', () => {
        if (!requestBody.length) {
          reject(null);
        } else {
          req.rawBody = Buffer.concat(requestBody);
          resolve(req);
        }
      });
      request.once('error', () => reject(null));
    } else {
      resolve(req);
    }
  });
}

function collectRes(request: http.ClientRequest): Promise<ResInfo> {
  return new Promise((resolve, reject) => {
    request.prependOnceListener('response', (response) => {
      const res: ResInfo = {
        status: response.statusCode ?? 0,
        headers: response.headers,
        type: hasbody(response.headers)
          ? typeis.is(response.headers['content-type'] ?? '', CONTENT_TYPES) || undefined
          : undefined,
        cost: 0,
      };
      if (res.type) {
        let input = new PassThrough();
        const decoders: Transform[] = [];
        const encodings = String(res.headers['content-encoding'] ?? '')
          .split(',')
          .map(val => val.trim())
          .filter(val => val.length);

        for (const encoding of encodings) {
          switch (encoding) {
            case 'gzip':
            case 'x-gzip':
            case 'compress':
            case 'x-compress':
            case 'deflate':
              decoders.push(zlib.createUnzip(zlibOptions));
              break;
            case 'br':
              decoders.push(zlib.createBrotliDecompress(brotliOptions));
              break;
          }
        }

        let timeout: any = setTimeout(() => resolve(res), 1000);
        const push = response.push;
        response.push = function(...args) {
          if (timeout !== null) {
            clearTimeout(timeout);
            timeout = null;
          }
          input.push(...args);
          return push.apply(this, args);
        }
        response.prependOnceListener('end', () => {
          input.end();
        });
        for (let i = 0; i < decoders.length; i++) {
          (decoders[i - 1] ?? input).pipe(decoders[i]);
        }
        const output = decoders.length ? decoders[decoders.length - 1] : input;
        const responseBody: Buffer[] = [];
        output.on('data', (chunk) => {
          responseBody.push(chunk);
        });
        output.on('end', () => {
          res.rawBody = Buffer.concat(responseBody);
          destroy(input);
          for (const decoder of decoders) {
            destroy(decoder);
          }
          resolve(res);
        });
        output.on('error', () => {
          destroy(input);
          for (const decoder of decoders) {
            destroy(decoder);
          }
        });
      } else {
        resolve(res);
      }
    });
    request.once('error', (error) => {
      reject({ error });
    });
  });
}

http.request = inspectedMethod.bind(http, http.request);
http.get = inspectedMethod.bind(http, http.get);
https.request = inspectedMethod.bind(https, https.request);
https.get = inspectedMethod.bind(https, https.get);