import zlib from 'zlib';
import { Transform, PassThrough } from 'stream';
import { Agent } from 'undici';
import typeis from 'type-is';
import destroy from 'destroy';
import { emit, ReqInfo, ResInfo } from './index';
import { CONTENT_TYPES, brotliOptions, zlibOptions } from './contant';
import { hasbody } from './utils';

export const inspectorAgent = withInspector(new Agent());

export function withInspector<T extends Agent>(agent: T): T {
  const dispatch = agent.dispatch;
  agent.dispatch = function(options, handler) {

    const reqHeaders:ReqInfo['headers'] = {};
    if (options.headers) {
      if (Array.isArray(options.headers)) {
        for (let n = 0; n < options.headers.length; n += 2) {
          const key = options.headers[n + 0].toLowerCase();
          const val = options.headers[n + 1].toLowerCase();
          reqHeaders[key] = val;
        }
      } else {
        Object.assign(reqHeaders, options.headers);
      }
    }
    const req: ReqInfo = {
      headers: reqHeaders,
      method: options.method,
      remoteAddress: '',
      time: Date.now(),
      type: options.body
        ? (typeis.is(String(reqHeaders['content-type'] ?? ''), CONTENT_TYPES) || undefined) as ReqInfo['type']
        : undefined,
      url: options.origin + options.path,
      rawBody: typeof options.body === 'string'
      ? Buffer.from(options.body)
      : Buffer.isBuffer(options.body)
        ? options.body
        : undefined,
      textBody: typeof options.body === 'string'
        ? options.body
        : Buffer.isBuffer(options.body)
          ? options.body.toString('utf-8')
          : undefined
    };

    const res: ResInfo = {
      status: 0,
      cost: 0,
      headers: {},
      type: undefined,
    };


    const input = new PassThrough();
    const decoders: Transform[] = [];
    const onHeaders = handler.onHeaders;
    handler.onHeaders = function(status, headersList, resume) {
      res.status = status;
      if (headersList) {
        for (let n = 0; n < headersList.length; n += 2) {
          const key = headersList[n + 0].toString('latin1').toLowerCase();
          const val = headersList[n + 1].toString('latin1').toLowerCase();
          res.headers[key] = val;
        }
      }
      res.type = hasbody(res.headers)
        ? (typeis.is(String(res.headers['content-type'] ?? ''), CONTENT_TYPES) || undefined) as ResInfo['type']
        : undefined;

      const encodings = String(res.headers['content-encoding'] ?? '')
        .split(',')
        .map(val => val.trim())
        .filter(val => val.length);

      for (const encoding of encodings) {
        switch (encoding) {
          case 'gzip':
          case 'x-gzip':
          case 'compress':
          case 'x-compress':
          case 'deflate':
            decoders.push(zlib.createUnzip(zlibOptions));
            break;
          case 'br':
            decoders.push(zlib.createBrotliDecompress(brotliOptions));
            break;
        }
      }

      const responseBody: Buffer[] = [];
      for (let i = 0; i < decoders.length; i++) {
        (decoders[i - 1] ?? input).pipe(decoders[i]);
      }
      const output = decoders.length ? decoders[decoders.length - 1] : input;
      output.on('data', (chunk) => {
        responseBody.push(chunk);
      });
      output.on('end', () => {
        res.cost = Date.now() - req.time;
        res.rawBody = Buffer.concat(responseBody);
        if (res.type) {
          res.textBody = res.rawBody.toString('utf-8');
        }
        destroy(input);
        for (const decoder of decoders) {
          destroy(decoder);
        }
        emit({ type: 'undici', req, res });
      });
      output.on('error', () => {
        destroy(input);
        for (const decoder of decoders) {
          destroy(decoder);
        }
      });

      if (onHeaders) {
        return onHeaders.call(this, status, headersList, resume);
      }
      return true;
    }

    const onData = handler.onData;
    handler.onData = function(chunk) {
      input.push(chunk);
      if (onData) {
        return onData.call(this, chunk);
      }
      return true;
    }

    const onComplete = handler.onComplete;
    handler.onComplete = function(trailers) {
      input.end();
      if (onComplete) {
        return onComplete.call(this, trailers);
      }
    }

    return dispatch.call(this, options, handler);
  }
  return agent;
}
